(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory SemanticLabeling imports DpFramework
begin

fun map_term :: "('f \<Rightarrow> 'g) \<Rightarrow> ('f,'v)term \<Rightarrow> ('g,'v)term"
where "map_term fg (Var x) = (Var x)"
    | "map_term fg (Fun f ts) = Fun (fg f) (map (map_term fg) ts)"

fun map_subst :: "('f \<Rightarrow> 'g) \<Rightarrow> ('f,'v)sub \<Rightarrow> ('g,'v)sub"
where "map_subst fg (Sub sigma) = Sub (\<lambda> x. map_term fg (sigma x))"

fun map_rule :: "('f \<Rightarrow> 'g) \<Rightarrow> ('f,'v)rule \<Rightarrow> ('g,'v)rule"
where "map_rule fg lr = (map_term fg (fst lr), map_term fg (snd lr))"

fun map_ctxt :: "('f \<Rightarrow> 'g) \<Rightarrow> ('f,'v)ctxt \<Rightarrow> ('g,'v)ctxt"
where "map_ctxt fg \<box> = \<box>"
   |  "map_ctxt fg (More f ss C ts) = More (fg f) (map (map_term fg) ss) (map_ctxt fg C) (map (map_term fg) ts)"

fun map_trs :: "('f \<Rightarrow> 'g) \<Rightarrow> ('f,'v)trel \<Rightarrow> ('g,'v)trel"
where "map_trs fg R = (map_rule fg) ` R" (* perhaps - {lr. fst lr = snd lr} to remove decr-rules *)

fun map_trsI :: "('f \<Rightarrow> 'g) \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('g,'v)trsL"
  where "map_trsI fg R = map (map_rule fg) R"

lemma map_trsI_sound[simp]: "set (map_trsI fg R) = map_trs fg (set R)" by auto
declare map_trsI.simps[simp del]


lemma map_term_comp : "map_term fg (map_term gh t) = map_term (fg o gh) t"
  by (induct t, auto simp: mapMap)

lemma map_trs_comp : "map_trs fg (map_trs gh R) = map_trs (fg o gh) R"
proof -  
  have mr: "map_rule (fg o gh) = map_rule fg o map_rule gh"
    by (rule ext, auto simp: map_term_comp)
  thus ?thesis
    by (auto simp: map_term_comp image_compose[symmetric] mr)
qed

lemma map_subst_distrib[simp] : "map_term fg (t \<cdot> \<sigma>) = map_term fg t \<cdot> map_subst fg \<sigma>"
proof (cases \<sigma>)
  case (Sub sigma)
  have "map_term fg (t \<cdot> Sub sigma) = map_term fg t \<cdot> map_subst fg (Sub sigma)"    
    by (induct t, auto simp: mapMap)
  with Sub show ?thesis by simp
qed

lemma map_ctxt_distrib[simp] : "map_term fg (C\<langle>t\<rangle>) = (map_ctxt fg C)\<langle>map_term fg t\<rangle>" 
  by  (induct C, auto)


lemma map_trs_mono : assumes "R \<subseteq> R'" shows "map_trs fg R \<subseteq> map_trs fg R'"
  using assms by auto


lemma map_trs_power_mono : 
  fixes R R' :: "('f,'v)trel" and fg :: "'f \<Rightarrow> 'f"
  assumes "R \<subseteq> R'" shows "((map_trs fg)^n) R \<subseteq> ((map_trs fg)^n) R'"
using assms by (induct n, simp, auto)

lemma rstep_map_trs_power_mono : 
  fixes R R' :: "('f,'v)trel" and fg :: "'f \<Rightarrow> 'f"
  assumes subset: "R \<subseteq> R'" shows "rstep (((map_trs fg)^n) R) \<subseteq> rstep (((map_trs fg)^n) R')"
  by (rule rstep_mono, rule map_trs_power_mono, rule subset)


lemma rewrite_imp_map_rewrite : assumes "(s,t) \<in> rstep R" 
 shows "(map_term fg s, map_term fg t) \<in> rstep (map_trs fg R)"
using assms
proof (induct)
  fix ss ts C
  assume "(ss,ts) \<in> CS R"
  thus "(map_term fg C\<langle>ss\<rangle>, map_term fg C\<langle>ts\<rangle>) \<in> rstep (map_trs fg R)" 
  proof
    fix s t \<sigma> 
    assume s: "ss = s \<cdot> \<sigma>" and t: "ts = t \<cdot> \<sigma>" and rule: "(s,t) \<in> R"
    from rule have lrule: "(map_term fg s, map_term fg t) \<in> map_trs fg R" by force
    from s t have lst: "map_term fg ss = map_term fg s \<cdot> map_subst fg \<sigma> \<and> map_term fg ts = map_term fg t \<cdot> map_subst fg \<sigma>" by auto
    with lrule have "(map_term fg ss, map_term fg ts) \<in> CS (map_trs fg R)" by auto
    thus "(map_term fg C\<langle>ss\<rangle>, map_term fg C\<langle>ts\<rangle>) \<in> rstep (map_trs fg R)" by auto
  qed
qed  

lemma rewrites_imp_map_rewrites : assumes "(s,t) \<in> (rstep R)^*"
  shows "(map_term fg s, map_term fg t) \<in> (rstep (map_trs fg R))^*"
using assms
proof (induct, simp)
  case (step y z)
  hence "(map_term fg y, map_term fg z) \<in> rstep (map_trs fg R)" using rewrite_imp_map_rewrite by auto
  with step show ?case by auto
qed


lemma SN_map_imp_SN : 
  fixes R:: "('g,'v)trel" and t :: "('g,'v)term"
  assumes SN: "SN (rstep (map_trs fg R)) (map_term fg t)" 
  shows "SN (rstep R) t"
proof (rule ccontr)
  assume "\<not> SN (rstep R) t"
  from this obtain f where cond: "f 0 = t \<and> (\<forall> i. (f i, f (Suc i)) \<in> rstep R)" 
  unfolding SN_elt_def by auto
  obtain g where g: "g = (\<lambda> i. map_term fg (f i))" by auto
  with cond have cond2: "g 0 = map_term fg t \<and> (\<forall> i. (g i, g (Suc i)) \<in> rstep (map_trs fg R))"
    using rewrite_imp_map_rewrite[where fg = fg] by auto
  from SN have "\<not> (\<exists> g. (g 0 = map_term fg t \<and> (\<forall> i. (g i, g (Suc i)) \<in> rstep (map_trs fg R))))" 
    unfolding SN_elt_def by auto
  with cond2 show False  by auto
qed

lemma ichain_imp_map_ichain : assumes chain: "ichain (P,R) s t \<sigma>"
  shows "ichain (map_trs fg P, map_trs fg R) (\<lambda> i. map_term fg (s i)) (\<lambda> i. map_term fg (t i)) (\<lambda> i. map_subst fg (\<sigma> i))"
  (is "ichain (?lP, ?lR) ?ls ?lt ?lsig")
proof -
  have mem: "\<forall> i. (?ls i, ?lt i) \<in> ?lP"
  proof
    fix i
    from chain have "(s i, t i) \<in> P" by auto
    thus "(?ls i, ?lt i) \<in> ?lP" by force
  qed
  have steps: "\<forall> i. (?lt i \<cdot> ?lsig i, ?ls (Suc i) \<cdot> ?lsig (Suc i)) \<in> (rstep ?lR)^*"
  proof
    fix i
    from chain have "(t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (rstep R)^*" by auto
    hence "(map_term fg (t i \<cdot> \<sigma> i), map_term fg (s (Suc i) \<cdot> \<sigma> (Suc i))) \<in> (rstep (map_trs fg R))^*" by (rule rewrites_imp_map_rewrites)
    thus "(?lt i \<cdot> ?lsig i, ?ls (Suc i) \<cdot> ?lsig (Suc i)) \<in> (rstep ?lR)^*" by auto
  qed
  from mem steps show ?thesis by auto
qed
  

text {* now we head for labelings *}

datatype ('f,'l)lab = Lab 'f "'l list"

(* unlab removes one layer of labelings *)
fun unlab :: "('f,'l)lab \<Rightarrow> ('f,'l)lab"
where "unlab (Lab f ls) = Lab f (tl ls)"

abbreviation unlab_term :: "(('f,'l)lab,'v)term \<Rightarrow> (('f,'l)lab,'v)term"
where "unlab_term \<equiv> map_term unlab"

abbreviation unlab_trs :: "(('f,'l)lab,'v)trel \<Rightarrow> (('f,'l)lab,'v)trel"
where "unlab_trs \<equiv> map_trs unlab"

abbreviation unlab_trsL :: "(('f,'l)lab,'v)trsL \<Rightarrow> (('f,'l)lab,'v)trsL"
where "unlab_trsL \<equiv> map (map_rule unlab)"

abbreviation unlab_subst :: "(('f,'l)lab,'v)sub \<Rightarrow> (('f,'l)lab,'v)sub"
where "unlab_subst \<equiv> map_subst unlab"

(* init_lab adds empty labelings to all function symbols *)
fun init_lab :: "'f \<Rightarrow> ('f,'l)lab"
where "init_lab f = Lab f []"

abbreviation init_lab_term :: "('f,'v)term \<Rightarrow> (('f,'l)lab,'v)term"
where "init_lab_term \<equiv> map_term init_lab"

abbreviation init_lab_subst :: "('f,'v)sub \<Rightarrow> (('f,'l)lab,'v)sub"
where "init_lab_subst \<equiv> map_subst init_lab"

abbreviation init_lab_rule :: "('f,'v)rule \<Rightarrow> (('f,'l)lab,'v)rule"
where "init_lab_rule \<equiv> map_rule init_lab"

abbreviation init_lab_trs :: "('f,'v)trel \<Rightarrow> (('f,'l)lab,'v)trel"
where "init_lab_trs \<equiv> map_trs init_lab"

abbreviation init_lab_trsL :: "('f,'v)trsL \<Rightarrow> (('f,'l)lab,'v)trsL"
where "init_lab_trsL \<equiv> map init_lab_rule"

(* remove_lab switches back to an unlabeled signature and thereby
   removes all labelings *)
fun remove_lab :: "('f,'l)lab \<Rightarrow> 'f" 
where "remove_lab (Lab f ls) = f"

lemma remove_lab_sound[simp] : "remove_lab (init_lab f) = f" by auto

abbreviation remove_lab_term :: "(('f,'l)lab,'v)term \<Rightarrow> ('f,'v)term"
where "remove_lab_term \<equiv> map_term remove_lab"

abbreviation remove_lab_subst :: "(('f,'l)lab,'v)sub \<Rightarrow> ('f,'v)sub"
where "remove_lab_subst \<equiv> map_subst remove_lab"


lemma remove_lab_term_sound : "remove_lab_term (init_lab_term t) = t"
proof (induct t, simp)
  case (Fun f ss)
  thus ?case by (induct ss, auto)
qed

abbreviation remove_lab_rule :: "(('f,'l)lab,'v)rule \<Rightarrow> ('f,'v)rule"
where "remove_lab_rule \<equiv> map_rule remove_lab"

lemma remove_lab_rule_sound : "remove_lab_rule (init_lab_rule lr) = lr"
  by (auto simp: remove_lab_term_sound)

abbreviation remove_lab_trs :: "(('f,'l)lab,'v)trel \<Rightarrow> ('f,'v)trel"
where "remove_lab_trs \<equiv> map_trs remove_lab"

lemma remove_lab_trs_sound : fixes R :: "('f,'v)trel"
  shows "remove_lab_trs (init_lab_trs R) = R"
proof 
  show "R \<subseteq> remove_lab_trs (init_lab_trs R :: (('f,'l)lab,'v)trel)" (is "_ \<subseteq> remove_lab_trs ?IR")
  proof 
    fix lr 
    assume "lr \<in> R"
    hence "remove_lab_rule ((init_lab_rule lr) :: (('f,'l)lab,'v)rule) \<in> (remove_lab_trs ?IR)" by force
    thus "lr \<in> remove_lab_trs ?IR" by (simp only: remove_lab_rule_sound)
  qed
next
  show "remove_lab_trs (init_lab_trs R) \<subseteq> R" by (auto simp: remove_lab_term_sound)
qed

definition lab_minimal_cond :: "(('f,'l)lab,'v)trel \<Rightarrow> (nat \<Rightarrow> (('f,'l)lab,'v)term) \<Rightarrow> (nat \<Rightarrow> (('f,'l)lab,'v)term) \<Rightarrow> (nat \<Rightarrow> (('f,'l)lab,'v)sub) \<Rightarrow> bool"
where "lab_minimal_cond R s t \<sigma> \<equiv> \<forall> i. (SN (rstep (remove_lab_trs R)) (remove_lab_term (t i \<cdot> \<sigma> i)))"

lemma lab_minimal_cond_imp_minimal_cond[simp] : "lab_minimal_cond R s t \<sigma> \<Longrightarrow> minimal_cond R s t \<sigma>"
proof -
  assume "lab_minimal_cond R s t \<sigma>"
  thus ?thesis using SN_map_imp_SN by (unfold lab_minimal_cond_def minimal_cond_def, blast)
qed

lemma lab_minimal_cond_mono : 
  assumes subset: "R \<subseteq> R'" and cond: "lab_minimal_cond R' s t \<sigma>" 
  shows "lab_minimal_cond R s t \<sigma>"
unfolding lab_minimal_cond_def
proof (rule allI)
  fix i
  from cond have SN: "SN (rstep (remove_lab_trs R')) (remove_lab_term (t i \<cdot> \<sigma> i))"
    unfolding lab_minimal_cond_def by auto
  have "rstep (remove_lab_trs R) \<subseteq> rstep (remove_lab_trs R')" by (rule rstep_mono, rule map_trs_mono, rule subset)
  with SN show "SN (rstep (remove_lab_trs R)) (remove_lab_term (t i \<cdot> \<sigma> i))" by (rule SN_imp_SN_subset)
qed

fun min_lab_ichain ::
 "(('f,'l)lab,'v)dpp \<Rightarrow> (nat \<Rightarrow> (('f,'l)lab,'v)term) \<Rightarrow> (nat \<Rightarrow> (('f,'l)lab,'v)term) \<Rightarrow> (nat \<Rightarrow> (('f,'l)lab,'v)sub) \<Rightarrow> bool"
where "min_lab_ichain (P,R) s t \<sigma> = (ichain (P,R) s t \<sigma> \<and> lab_minimal_cond R s t \<sigma>)"

lemma min_lab_ichainI[intro] : 
  assumes "R \<subseteq> R'" and "lab_minimal_cond R' s t \<sigma>" and "ichain (P,R) (shift s i) (shift t i) (shift \<sigma> i)" 
  shows "min_lab_ichain (P,R) (shift s i) (shift t i) (shift \<sigma> i)"
proof -
  from assms have "lab_minimal_cond R s t \<sigma>" using lab_minimal_cond_mono by auto
  hence "lab_minimal_cond R (shift s i) (shift t i) (shift \<sigma> i)" 
    unfolding lab_minimal_cond_def by auto
  with assms show ?thesis by auto
qed


definition finite_lab_dpp :: "(('f,'l)lab,'v)dpp \<Rightarrow> bool" where "finite_lab_dpp DPP = (\<not>(\<exists>s t \<sigma>. min_lab_ichain DPP s t \<sigma>))"

lemma init_lab_sound: assumes finite: "finite_lab_dpp (init_lab_trs P, init_lab_trs R)"
  shows "finite_dpp (P,R)"
proof (rule ccontr)
  assume "\<not> finite_dpp (P,R)"
  from this obtain s t \<sigma> where chain: "min_ichain (P,R) s t \<sigma>" unfolding finite_dpp_def by auto
  hence ichain: "ichain (P,R) s t \<sigma>" by auto
  let ?ls = "\<lambda> i. init_lab_term (s i)"
  let ?lt = "\<lambda> i. init_lab_term (t i)"
  let ?lsig = "\<lambda> i. init_lab_subst (\<sigma> i)" 
  let ?lP = "init_lab_trs P"
  let ?lR = "init_lab_trs R"

  have ichain2: "ichain (?lP,?lR) ?ls ?lt ?lsig" 
    by (rule ichain_imp_map_ichain, rule ichain)
  
  have "\<forall> i. SN (rstep (remove_lab_trs ?lR)) (remove_lab_term (?lt i \<cdot> ?lsig i))"
  proof
    fix i
    from chain have "SN (rstep R) (t i \<cdot> \<sigma> i)" by (auto simp: minimal_cond_def)
    hence "SN (rstep (remove_lab_trs ?lR)) (remove_lab_term (init_lab_term (t i \<cdot> \<sigma> i)))" 
      by (simp only: remove_lab_trs_sound remove_lab_term_sound)
    thus "SN (rstep (remove_lab_trs ?lR)) (remove_lab_term (init_lab_term (t i) \<cdot> init_lab_subst (\<sigma> i)))" by simp
  qed

  with ichain2 have "min_lab_ichain (?lP,?lR) ?ls ?lt ?lsig" by (auto simp: lab_minimal_cond_def)

  with finite show False unfolding finite_lab_dpp_def by blast
qed

lemma remove_unlab_is_remove : "remove_lab o unlab = remove_lab"
proof (rule ext)
  fix fl
  show "(remove_lab o unlab) fl = remove_lab fl" by (cases fl, auto)
qed



lemma unlab_proc_sound : fixes R :: "(('f,'l)lab,'v)trel"
  assumes finite: "finite_lab_dpp (unlab_trs P, unlab_trs R)"
  shows "finite_lab_dpp (P,R)"
proof (rule ccontr)
  assume "\<not> finite_lab_dpp (P,R)"
  from this obtain s t \<sigma> where chain: "min_lab_ichain (P,R) s t \<sigma>" unfolding finite_lab_dpp_def by auto
  hence ichain: "ichain (P,R) s t \<sigma>" by auto
  let ?ls = "\<lambda> i. unlab_term (s i)"
  let ?lt = "\<lambda> i. unlab_term (t i)"
  let ?lsig = "\<lambda> i. unlab_subst (\<sigma> i)" 
  let ?lP = "unlab_trs P"
  let ?lR = "unlab_trs R"
  from ichain have ichain2: "ichain (?lP, ?lR) ?ls ?lt ?lsig" using ichain_imp_map_ichain by blast
  
  have SN: "\<forall> i. SN (rstep (remove_lab_trs ?lR)) (remove_lab_term (?lt i \<cdot> ?lsig i))"
  proof 
    fix i
    from chain have SN: "SN (rstep (remove_lab_trs R)) (remove_lab_term (t i \<cdot> \<sigma> i))" by (auto simp: lab_minimal_cond_def)
     have id1: "remove_lab_trs R = remove_lab_trs ?lR" by (simp only: map_trs_comp remove_unlab_is_remove)
    have "remove_lab_term (?lt i \<cdot> ?lsig i) = remove_lab_term (unlab_term (t i \<cdot> \<sigma> i))" by auto
    also have "\<dots> = remove_lab_term (t i \<cdot> \<sigma> i)" by (simp only: map_term_comp remove_unlab_is_remove) 
    finally show "SN (rstep (remove_lab_trs ?lR)) (remove_lab_term (?lt i \<cdot> ?lsig i))" using id1 SN by auto
  qed

  from ichain2 SN have "min_lab_ichain (?lP, ?lR) ?ls ?lt ?lsig" by (auto simp: lab_minimal_cond_def)
  with finite show False unfolding finite_lab_dpp_def by blast
qed

lemma remove_lab_proc_sound : 
  assumes finite: "finite_dpp (remove_lab_trs P, remove_lab_trs R)"
  shows "finite_lab_dpp (P,R)"
proof (rule ccontr)
  assume "\<not> finite_lab_dpp (P,R)"
  from this obtain s t \<sigma> where chain: "min_lab_ichain (P,R) s t \<sigma>" unfolding finite_lab_dpp_def by auto
  hence ichain: "ichain (P,R) s t \<sigma>" by auto
  let ?ls = "\<lambda> i. remove_lab_term (s i)"
  let ?lt = "\<lambda> i. remove_lab_term (t i)"
  let ?lsig = "\<lambda> i. remove_lab_subst (\<sigma> i)" 
  let ?lP = "remove_lab_trs P"
  let ?lR = "remove_lab_trs R"
  from ichain have ichain2: "ichain (?lP, ?lR) ?ls ?lt ?lsig" using ichain_imp_map_ichain by blast
  
  have "minimal_cond ?lR ?ls ?lt ?lsig" using chain by (auto simp: lab_minimal_cond_def minimal_cond_def)
  with ichain2 have "min_ichain (?lP, ?lR) ?ls ?lt ?lsig" by simp
  with finite show False unfolding finite_dpp_def by blast
qed

definition sound_lab_proc 
  where "sound_lab_proc proc \<equiv> \<forall> P R rP rR. (proc P R rP rR \<longrightarrow> (finite_lab_dpp (rP, rR) \<longrightarrow> finite_lab_dpp (P, R)))"

definition sound_lab_proc_impl
  where "sound_lab_proc_impl proc \<equiv> \<forall> P R rP rR. (proc P R (computeRuleMap R) rP rR \<longrightarrow> (finite_lab_dpp (set rP, set rR) \<longrightarrow> finite_lab_dpp (set P, set R)))"

definition sound_lab_onlyDP_proc_impl
  where "sound_lab_onlyDP_proc_impl proc \<equiv> \<forall> P R rP. (proc P R (computeRuleMap R) rP \<longrightarrow> (finite_lab_dpp (set rP, set R) \<longrightarrow> finite_lab_dpp (set P, set R)))"


lemma ci_proc_to_sound_lab_proc : assumes ci_proc: "chain_identifying_proc proc"
  shows "sound_lab_proc proc"
unfolding sound_lab_proc_def
proof (intro allI, intro impI)
  fix P R rP rR
  assume cond: "proc P R rP rR" and finite: "finite_lab_dpp (rP, rR)"
  show "finite_lab_dpp (P,R)"
  proof (rule ccontr)
    assume "\<not> finite_lab_dpp (P,R)"
    from this obtain s t \<sigma> where ml: "min_lab_ichain (P,R) s t \<sigma>" unfolding finite_lab_dpp_def by auto
    hence "min_ichain (P,R) s t \<sigma>" by auto
    with ci_proc cond obtain i where chain: "rR \<subseteq> R \<and> ichain (rP,rR) (shift s i) (shift t i) (shift \<sigma> i)" unfolding chain_identifying_proc_def by force
    have "min_lab_ichain (rP,rR) (shift s i) (shift t i) (shift \<sigma> i)" 
      using min_lab_ichainI[where R' = R and R = rR] chain ml by auto
    hence "\<not> finite_lab_dpp (rP,rR)" unfolding finite_lab_dpp_def by blast
    with finite show False by simp
  qed
qed 

lemma ci_proc_to_sound_lab_proc_impl[simp] : assumes ci_proc: "chain_identifying_proc_impl proc"
  shows "sound_lab_proc_impl proc"
unfolding sound_lab_proc_impl_def
proof (intro allI, intro impI)
  fix P R rP rR
  assume cond: "proc P R (computeRuleMap R) rP rR" and finite: "finite_lab_dpp (set rP, set rR)"
  show "finite_lab_dpp (set P, set R)"
  proof (rule ccontr)
    assume "\<not> finite_lab_dpp (set P,set R)"
    from this obtain s t \<sigma> where ml: "min_lab_ichain (set P,set R) s t \<sigma>" unfolding finite_lab_dpp_def by auto
    hence "min_ichain (set P,set R) s t \<sigma>" by auto
    with ci_proc cond obtain i where chain: "set rR \<subseteq> set R \<and> ichain (set rP,set rR) (shift s i) (shift t i) (shift \<sigma> i)" unfolding chain_identifying_proc_impl_def by force
    have "min_lab_ichain (set rP,set rR) (shift s i) (shift t i) (shift \<sigma> i)" 
      using min_lab_ichainI[where R' = "set R" and R = "set rR"] chain ml by auto
    hence "\<not> finite_lab_dpp (set rP,set rR)" unfolding finite_lab_dpp_def by blast
    with finite show False by simp
  qed
qed 

lemma ci_proc_to_sound_lab_onlyDP_proc_impl[simp] : assumes ci_proc: "chain_identifying_onlyDP_proc_impl proc"
  shows "sound_lab_onlyDP_proc_impl proc"
unfolding sound_lab_onlyDP_proc_impl_def
proof (intro allI, intro impI)
  fix P R rP 
  assume cond: "proc P R (computeRuleMap R) rP" and finite: "finite_lab_dpp (set rP, set R)"
  show "finite_lab_dpp (set P, set R)"
  proof (rule ccontr)
    assume "\<not> finite_lab_dpp (set P,set R)"
    from this obtain s t \<sigma> where ml: "min_lab_ichain (set P,set R) s t \<sigma>" unfolding finite_lab_dpp_def by auto
    hence "min_ichain (set P,set R) s t \<sigma>" by auto
    with ci_proc cond obtain i where chain: "ichain (set rP,set R) (shift s i) (shift t i) (shift \<sigma> i)" unfolding chain_identifying_onlyDP_proc_impl_def by blast
    have "min_lab_ichain (set rP,set R) (shift s i) (shift t i) (shift \<sigma> i)" 
      using min_lab_ichainI[where R' = "set R" and R = "set R"] chain ml by auto
    hence "\<not> finite_lab_dpp (set rP,set R)" unfolding finite_lab_dpp_def by blast
    with finite show False by simp
  qed
qed 

lemma apply_ci_onlyDP_lab_proc : assumes ci: "chain_identifying_onlyDP_proc_impl proc"
  and fin: "finite_lab_dpp (set rP, set R)"
  and proc: "proc P R (computeRuleMap R) rP"
  shows "finite_lab_dpp (set P, set R)"
proof -
  from ci have "sound_lab_onlyDP_proc_impl proc" by simp
  thus ?thesis unfolding sound_lab_onlyDP_proc_impl_def using fin proc by blast
qed

lemma apply_ci_lab_proc : assumes ci: "chain_identifying_proc_impl proc"
  and fin: "finite_lab_dpp (set rP, set rR)"
  and proc: "proc P R (computeRuleMap R) rP rR"
  shows "finite_lab_dpp (set P, set R)"
proof -
  from ci have "sound_lab_proc_impl proc" by simp
  thus ?thesis unfolding sound_lab_proc_impl_def using fin proc by blast
qed

lemma apply_sound_lab_proc : assumes sound: "sound_lab_proc_impl proc"
  and fin: "finite_lab_dpp (set rP, set rR)"
  and proc: "proc P R (computeRuleMap R) rP rR"
  shows "finite_lab_dpp (set P, set R)"
proof -
  from sound show ?thesis unfolding sound_lab_proc_impl_def using fin proc by blast
qed


text {* Now let us head for the semantic part *}

types ('v,'c)assign = "'v \<Rightarrow> 'c"
types ('f,'c,'l)inter = "('f,'l)lab \<Rightarrow> 'c list \<Rightarrow> 'c"
types ('f,'l)labels = "('f,'l)lab \<Rightarrow> nat \<Rightarrow> 'l set"
types ('f,'c,'l)label = "('f,'l)lab \<Rightarrow> 'c list \<Rightarrow> 'l"

fun label :: "('f,'l)lab \<Rightarrow> 'l \<Rightarrow> ('f,'l)lab"
where "label (Lab f ls) l = Lab f (l # ls)"

fun eval_lab :: "('f,'c,'l)inter \<Rightarrow> ('f,'c,'l)label \<Rightarrow> ('v,'c)assign \<Rightarrow> (('f,'l)lab,'v)term \<Rightarrow> ('c \<times> (('f,'l)lab,'v)term)" 
where "eval_lab I L \<alpha> (Var x) = (\<alpha> x, Var x)"
  |   "eval_lab I L \<alpha> (Fun f ts) = 
     (let clts = map (eval_lab I L \<alpha>) ts in 
     (let cs = map fst clts in 
     (let c = I f cs in
     (let lts = map snd clts in 
     (let l = L f cs in
     (c, Fun (label f l) lts))))))"

abbreviation eval where "eval I L \<alpha> t \<equiv> fst (eval_lab I L \<alpha> t)"
abbreviation lab where "lab I L \<alpha> t \<equiv> snd (eval_lab I L \<alpha> t)"
definition wf_assign where "wf_assign C \<alpha> \<equiv> \<forall> x. \<alpha> x \<in> C"
definition lab_rule where "lab_rule I L C lr \<equiv> {(lab I L \<alpha> (fst lr), lab I L \<alpha> (snd lr)) | \<alpha> . wf_assign C \<alpha>}"
abbreviation lab_trs where "lab_trs I L C R \<equiv> \<Union>(lab_rule I L C ` R)"
definition undo_lab where "undo_lab f \<equiv> (fst f, tl (snd f))"
definition wf_inter where "wf_inter I C \<equiv> \<forall> f cs. set cs \<subseteq> C \<longrightarrow> I f cs \<in> C"
abbreviation model_rule where "model_rule I L C l r \<equiv> \<forall> \<alpha>. wf_assign C \<alpha> \<longrightarrow> eval I L \<alpha> l  = eval I L \<alpha> r"
definition model where "model I L C R \<equiv> \<forall> (l,r) \<in> R. model_rule I L C l r"


locale sl_interpr = 
  fixes C :: "('c :: sshow) set"
  and  c :: "'c"
  and  I :: "('f :: sshow,'c :: sshow,'l :: sshow)inter"
(*  and  LS :: "('f,'l)labels" *)
  and  L :: "('f,'c,'l)label"
  assumes c: "c \<in> C"
  and     wf_I: "wf_inter I C"
begin

abbreviation Eval where "Eval \<equiv> eval I L"
abbreviation Lab where "Lab \<equiv> lab I L"
abbreviation Lab_trs :: "(('f,'l)lab,'v)trel \<Rightarrow> (('f,'l)lab,'v)trel" where "Lab_trs \<equiv> lab_trs I L C"
abbreviation wf_ass where "wf_ass \<equiv> wf_assign C"
definition sub_ass where "sub_ass \<alpha> \<sigma> \<equiv> \<lambda> x. Eval \<alpha> (getSub \<sigma> x)"
definition lab_sub where "lab_sub \<alpha> \<sigma> \<equiv> Sub (\<lambda> x. Lab \<alpha> (getSub \<sigma> x))"

lemma [simp]: "unlab (label f l) = f" by (cases f, auto)

lemma unlab_term: "unlab_term (lab I L \<alpha> t) = t"
proof (induct t, simp)
  case (Fun f ts)
  thus ?case by (auto simp: Let_def mapMap, induct ts, auto)
qed
lemma [simp]: "remove_lab (label f l) = remove_lab f" by (cases f, auto)

definition default_ass where "default_ass x = c"

lemma wf_default_ass: "wf_ass default_ass" 
  unfolding default_ass_def wf_assign_def using c by auto

abbreviation LAB where "LAB t \<equiv> Lab default_ass t"

lemma unlab_rule[simp]: 
   "unlab_trs (lab_rule I L C rule) = {rule}" (is "?rule = {rule}")
proof 
  show "?rule \<subseteq> {rule}"
  proof
    fix ulrule
    assume ulrule: "ulrule \<in> ?rule"
    show "ulrule \<in> {rule}"
    proof (cases ulrule)
      case (Pair ull ulr)
      with ulrule have "(ull,ulr) \<in> ?rule" by auto
      from this obtain ll lr where lrule: "ull = unlab_term ll \<and> ulr = unlab_term lr \<and> (ll,lr) \<in> lab_rule I L C rule" 
	by force
      from this obtain l r \<alpha> where "rule = (l,r) \<and> ll = lab I L \<alpha> l \<and> lr = lab I L \<alpha> r" unfolding lab_rule_def by (cases rule, auto)
      with Pair ulrule lrule show ?thesis by (simp add: unlab_term)
    qed
  qed
next
  show "{rule} \<subseteq> ?rule"
  proof (cases rule)
    case (Pair l r)
    with wf_default_ass
    have "(LAB l, LAB r) \<in> lab_rule I L C rule" (is "(?l,?r) \<in> _") 
      unfolding lab_rule_def using Pair by auto
    hence "(unlab_term ?l, unlab_term ?r) \<in> ?rule"  by force
    hence "rule \<in> ?rule" using Pair by (simp add: unlab_term)
    thus ?thesis by auto
  qed
qed



lemma remove_label_term[simp]: "remove_lab_term (Lab \<alpha> t) = remove_lab_term t"
proof (induct t, simp)
  case (Fun f ts)
  thus ?case by (auto simp: Let_def mapMap)
qed  

lemma remove_label_rule: "remove_lab_trs (lab_rule I L C rule) = {remove_lab_rule rule}" (is "?rule = {?rrule}")
proof 
  show "?rule \<subseteq> {?rrule}"
  proof
    fix ulrule
    assume ulrule: "ulrule \<in> ?rule"
    show "ulrule \<in> {?rrule}"
    proof (cases ulrule)
      case (Pair ull ulr)
      with ulrule have "(ull,ulr) \<in> ?rule" by auto
      from this obtain ll lr where lrule: "ull = remove_lab_term ll \<and> ulr = remove_lab_term lr \<and> (ll,lr) \<in> lab_rule I L C rule" 
	by force
      from this obtain l r \<alpha> where "rule = (l,r) \<and> ll = Lab \<alpha> l \<and> lr = Lab \<alpha> r" unfolding lab_rule_def by (cases rule, auto)
      with Pair ulrule lrule show ?thesis by (simp)
    qed
  qed
next
  show "{?rrule} \<subseteq> ?rule"
  proof (cases rule)
    case (Pair l r)
    with wf_default_ass have "(LAB l, LAB r) \<in> lab_rule I L C rule" unfolding lab_rule_def by auto    
    hence "(remove_lab_term (LAB l), remove_lab_term (LAB r)) \<in> ?rule"  by force
    hence "?rrule \<in> ?rule" using Pair by (simp)
    thus ?thesis by auto
  qed
qed

lemma remove_label_trs: fixes R :: "(('f,'l)lab,'v)trel"
  shows "remove_lab_trs (Lab_trs R) = remove_lab_trs R"
proof (rule set_ext)
  fix rule :: "('f,'v)rule"
  let ?R = "remove_lab_trs R :: ('f,'v)trel"
  let ?ULR = "remove_lab_trs (Lab_trs R) :: ('f,'v)trel"
  show "(rule \<in> ?ULR) = (rule \<in> ?R)"
  proof
    assume rule: "rule \<in> ?ULR"
    from this obtain rrule where cond: "rrule \<in> R" and "rule \<in> remove_lab_trs (lab_rule I L C rrule)" by auto
    hence "rule = remove_lab_rule rrule" using remove_label_rule[where ?rule = rrule] by auto
    thus "rule \<in> ?R" using cond map_trs_mono by auto
  next
    assume rule: "rule \<in> ?R"
    from this obtain rrule where cond: "rrule \<in> R" and "rule = remove_lab_rule rrule" by auto
    with remove_label_rule[where ?rule = rrule] have "rule \<in> remove_lab_trs (lab_rule I L C rrule)" by auto  
    with map_trs_mono rule cond show "rule \<in> ?ULR" by auto
  qed
qed

lemma remove_label_trs_2: "remove_lab_trs (unlab_trs R) = remove_lab_trs R"
proof -
    show ?thesis by (simp only: map_trs_comp, force simp: remove_unlab_is_remove)
qed


lemma unlab_trs: "unlab_trs (Lab_trs R) = R"
proof (rule set_ext)
  fix rule
  let ?ULR = "unlab_trs (Lab_trs R)"
  show "(rule \<in> ?ULR) = (rule \<in> R)"
  proof
    assume rule: "rule \<in> R"
    from unlab_rule[where ?rule = rule] have "rule \<in> unlab_trs (lab_rule I L C rule)" by auto  
    with map_trs_mono rule show "rule \<in> ?ULR" by auto
  next
    assume rule: "rule \<in> ?ULR"
    from this obtain rrule where "rrule \<in> R \<and> rule \<in> unlab_trs (lab_rule I L C rrule)" by auto
    thus "rule \<in> R" using rule unlab_rule[where ?rule = rrule] by auto
  qed
qed


lemma SN_imp_SN_lab_term: fixes t :: "(('f,'l)lab,'v)term" and R :: "(('f,'l)lab,'v)trel"
  assumes SN: "SN (rstep R) t"
  shows "SN (rstep (Lab_trs R)) (LAB t)"
proof -
  from SN have "SN (rstep (unlab_trs (Lab_trs R))) (unlab_term (LAB t))" 
    by (simp only: unlab_trs unlab_term)
  thus ?thesis by (rule SN_map_imp_SN)
qed    

lemma SN_imp_SN_lab_term_2: fixes t :: "(('f,'l)lab,'v)term" and R :: "(('f,'l)lab,'v)trel"
  assumes SN: "SN (rstep R) t"
  and subset: "unlab_trs LR \<subseteq> R"
  shows "SN (rstep LR) (LAB t)"
proof -
  from SN have "SN (rstep R) (unlab_term (LAB t))" 
    by (simp only: unlab_trs unlab_term)  
  show ?thesis 
    by (rule SN_map_imp_SN[where fg = unlab], simp only: unlab_term, rule SN_imp_SN_subset, rule SN, rule rstep_mono, rule subset)
qed


lemma eval_sub[simp] :
  "Eval \<alpha> (t \<cdot> \<sigma>) = Eval (sub_ass \<alpha> \<sigma>) t"
proof (induct t, cases \<sigma>, simp add: sub_ass_def)
  case (Fun f ts)
  let ?beta = "sub_ass \<alpha> \<sigma>"
  have evalInd: "map (\<lambda> x. Eval \<alpha> (x \<cdot> \<sigma>)) ts = map (Eval ?beta) ts" using Fun  by auto
  have "Eval \<alpha> (Fun f ts \<cdot> \<sigma>) = Eval \<alpha> (Fun f (map (\<lambda> t. t \<cdot> \<sigma>) ts))" by simp
  also have "\<dots> = Eval ?beta (Fun f ts)" by (simp add: Let_def mapMap evalInd)
  finally show ?case .
qed

lemma lab_sub[simp] :
  "Lab \<alpha> (t \<cdot> \<sigma>) = Lab (sub_ass \<alpha> \<sigma>) t \<cdot> lab_sub \<alpha> \<sigma>"
  by (induct t, cases \<sigma>, simp add: lab_sub_def, simp add: Let_def mapMap)
    

lemma wf_term[simp] : assumes "wf_ass \<alpha>"
  shows "Eval \<alpha> t \<in> C"
proof (induct t)
  case (Var x)
  with assms show ?case by (auto simp: wf_assign_def)
next
  case (Fun f ts)
  let ?cs = "map (Eval \<alpha>) ts"
  let ?c = "I f ?cs"
  from Fun have "set ?cs \<subseteq> C" by auto
  with wf_I have "?c \<in> C" unfolding wf_inter_def by (best)
  thus ?case by (simp add: Let_def mapMap)
qed

lemma wf_ass_sub_ass[simp] : assumes "wf_ass \<alpha>"
  shows "wf_ass (sub_ass \<alpha> \<sigma>)"
using assms unfolding sub_ass_def by (simp add: wf_assign_def)


lemma sem_rewrite: assumes step: "(s,t) \<in> rstep R"
  and model: "model I L C R"
  and wfass: "wf_ass \<alpha>"
  shows "Eval \<alpha> s = Eval \<alpha> t"
using step 
proof 
  fix ls rs C
  assume s: "s = C\<langle>ls\<rangle>" and t: "t = C\<langle>rs\<rangle>" and CS: "(ls,rs) \<in> CS R"
  have root: "Eval \<alpha> ls = Eval \<alpha> rs" using CS
  proof
    fix l r \<sigma>
    assume ls: "ls = l \<cdot> \<sigma>" and rs: "rs = r \<cdot> \<sigma>" and rule: "(l,r) \<in> R"
    let ?beta = "sub_ass \<alpha> \<sigma>"
    have wfbeta: "wf_ass ?beta" using wfass by auto
    from ls have "Eval \<alpha> ls = Eval ?beta l" by simp
    also have "\<dots> = Eval ?beta r" using rule wfbeta model unfolding model_def by best 
    also have "\<dots> = Eval \<alpha> rs" using rs by simp
    finally show ?thesis .
  qed
  have "Eval \<alpha> s = Eval \<alpha> (C\<langle>ls\<rangle>)" using s by simp
  also have "\<dots>  = Eval \<alpha> (C\<langle>rs\<rangle>)" 
    by (induct C, simp add: root, simp add: Let_def)
  also have "\<dots> = Eval \<alpha> t" using t by simp
  finally show ?thesis .
qed

lemma lab_rewrite: assumes step: "(s,t) \<in> rstep R"
  and model: "model I L C R"
  and wfass: "wf_ass \<alpha>"
  shows "(Lab \<alpha> s, Lab \<alpha> t) \<in> rstep (Lab_trs R)"
using step 
proof 
  fix ls rs C
  assume s: "s = C\<langle>ls\<rangle>" and t: "t = C\<langle>rs\<rangle>" and CS: "(ls,rs) \<in> CS R"
  have root: "(Lab \<alpha> ls, Lab \<alpha> rs) \<in> rstep (Lab_trs R)" using CS
  proof
    fix l r \<sigma>
    assume ls: "ls = l \<cdot> \<sigma>" and rs: "rs = r \<cdot> \<sigma>" and rule: "(l,r) \<in> R"
    let ?beta = "sub_ass \<alpha> \<sigma>"
    let ?delt = "lab_sub \<alpha> \<sigma>"
    have wfbeta: "wf_ass ?beta" using wfass by auto
    from ls rs have id: "(Lab \<alpha> ls, Lab \<alpha> rs) = (Lab ?beta l \<cdot> ?delt, Lab ?beta r \<cdot> ?delt)" by auto
    from rule wfbeta have "(Lab ?beta l, Lab ?beta r) \<in> (Lab_trs R)" by (simp add: lab_rule_def, force)
    hence "(Lab ?beta l \<cdot> ?delt, Lab ?beta r \<cdot> ?delt) \<in> CS (Lab_trs R)" by (rule csI)
    hence "(Lab ?beta l \<cdot> ?delt, Lab ?beta r \<cdot> ?delt) \<in> rstep (Lab_trs R)" using ccI[where C = \<box>] by force
    with id show ?thesis by simp
  qed   
  have id: "(Lab \<alpha> s, Lab \<alpha> t) = (Lab \<alpha> C\<langle>ls\<rangle>, Lab \<alpha> C\<langle>rs\<rangle>)" using s t by auto
  have "(ls, rs) \<in> rstep R" using CS ccI[where C = \<box>] by auto
  hence "Eval \<alpha> ls = Eval \<alpha> rs" using model wfass by (rule sem_rewrite)
  have "(Lab \<alpha> C\<langle>ls\<rangle>, Lab \<alpha> C\<langle>rs\<rangle>) \<in> rstep (Lab_trs R)" 
  proof (induct C)
    case Hole
    show ?case using root by simp
  next
    case (More f ss C ts)
    let ?D = "More (label f (L f (map (Eval \<alpha>) (ss @ C\<langle>ls\<rangle> # ts)))) (map (Lab \<alpha>) ss) \<box> (map (Lab \<alpha>) ts)"
    from CS have "(C\<langle>ls\<rangle>, C\<langle>rs\<rangle>) \<in> rstep R" using ccI[where C = C] by auto
    hence "Eval \<alpha> C\<langle>ls\<rangle> = Eval \<alpha> C\<langle>rs\<rangle>" using model wfass by (rule sem_rewrite)
    hence id: "(Lab \<alpha> More f ss C ts\<langle>ls\<rangle>,Lab \<alpha> More f ss C ts\<langle>rs\<rangle>) = (?D\<langle>Lab \<alpha> C\<langle>ls\<rangle>\<rangle>, ?D\<langle>Lab \<alpha> C\<langle>rs\<rangle>\<rangle>)" (is "_ = ?p2") 
      by (auto simp: Let_def mapMap)
    with More have "?p2 \<in> CC (rstep (Lab_trs R))" using ccI[where C = ?D] by force
    with ctxt_closed_rstep id show ?case unfolding ctxt_closed_def by auto
  qed
  with s t show ?thesis by auto 
qed 

lemma rewrites_imp_lab_rewrites: 
  assumes "(s,t) \<in> (rstep R)^*"
  and model: "model I L C R"
  shows "(LAB s, LAB t) \<in> (rstep (Lab_trs R))^*"
using assms
proof (induct, simp)
  case (step y z)
  with lab_rewrite[where s=y and t=z and \<alpha> = default_ass] model wf_default_ass show ?case by force
qed    
 
lemma SN_lab_imp_SN_term: fixes t :: "(('f,'l)lab,'v)term" and R :: "(('f,'l)lab,'v)trel"
  assumes SN: "SN (rstep (Lab_trs R)) (LAB t)" 
  and model: "model I L C R"
  shows "SN (rstep R) t"
proof (rule ccontr)
  assume "\<not> SN (rstep R) t"
  from this obtain f where inf: "f 0 = t \<and> (\<forall> i. (f i, f (Suc i)) \<in> rstep R)" unfolding SN_elt_def  by auto
  obtain g where g: "\<And> i. g i = LAB (f i)" by auto
  have one: "g 0 = LAB t" using inf g by auto
  have two: "\<forall> i. (g i, g (Suc i)) \<in> rstep (Lab_trs R)" 
    by (rule allI, simp only: g, rule lab_rewrite, auto simp: inf wf_default_ass model g)
  from one two SN show False unfolding SN_elt_def by auto
qed    


definition sem_lab_proc :: "
  (('f,'l)lab,'v)trel \<Rightarrow> 
  (('f,'l)lab,'v)trel \<Rightarrow> 
  (('f,'l)lab,'v)trel \<Rightarrow> 
  (('f,'l)lab,'v)trel \<Rightarrow> 
  bool"
  where "sem_lab_proc P R lP lR \<equiv> Lab_trs P \<subseteq> lP \<and> Lab_trs R \<subseteq> lR \<and> unlab_trs lR \<subseteq> R \<and> model I L C R"


definition sem_lab_trs :: "
  (('f,'l)lab,'v)trel \<Rightarrow> 
  (('f,'l)lab,'v)trel \<Rightarrow> 
  bool"
  where "sem_lab_trs R lR \<equiv> Lab_trs R \<subseteq> lR \<and> model I L C R"

abbreviation lab_terms where "lab_terms s \<sigma> i \<equiv> Lab (sub_ass default_ass (\<sigma> i)) (s i)"
abbreviation lab_subs  where "lab_subs  \<sigma> i \<equiv> lab_sub default_ass (\<sigma> i)"


lemma sem_lab_proc_sound : 
  shows "sound_lab_proc sem_lab_proc"
unfolding sound_lab_proc_def
proof (intro allI, intro impI, unfold sem_lab_proc_def)
  fix P R lP lR :: "(('f,'l)lab,'v)trel"
  assume "Lab_trs P \<subseteq> lP \<and> Lab_trs R \<subseteq> lR \<and> unlab_trs lR \<subseteq> R \<and> model I L C R" 
  hence model: "model I L C R" and lR1: "Lab_trs R \<subseteq> lR" and lR2: "unlab_trs lR \<subseteq> R" and lP: "Lab_trs P \<subseteq> lP" by auto
  let ?lP = "Lab_trs P"
  let ?lR = "Lab_trs R"
  assume finite: "finite_lab_dpp (lP, lR)"
  show "finite_lab_dpp (P,R)"
  proof (rule ccontr)
    assume "\<not> finite_lab_dpp (P,R)"
    from this obtain s t \<sigma> where chain: "min_lab_ichain (P,R) s t \<sigma>" unfolding finite_lab_dpp_def by auto

    let ?ls = "lab_terms s \<sigma>"
    let ?lt = "lab_terms t \<sigma>"
    let ?lsig = "lab_subs \<sigma>"
    let ?lss = "\<lambda> i. Lab default_ass (s i \<cdot> \<sigma> i)"
    let ?lts = "\<lambda> i. Lab default_ass (t i \<cdot> \<sigma> i)"

    have s: "\<forall> i. ?lss i = ?ls i \<cdot> ?lsig i" by simp
    have t: "\<forall> i. ?lts i = ?lt i \<cdot> ?lsig i" by simp

    have rstepSub: "(rstep ?lR)^* \<subseteq> (rstep lR)^*" 
      by (rule rtrancl_mono, rule rstep_mono, rule lR1)

    from chain have "\<forall> i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (rstep R)^*" by auto
    hence "\<forall> i. (?lts i, ?lss (Suc i)) \<in> (rstep ?lR)^*" using rewrites_imp_lab_rewrites model by best
    with s t have "\<forall> i. (?lt i \<cdot> ?lsig i, ?ls (Suc i) \<cdot> ?lsig (Suc i)) \<in> (rstep ?lR)^*" by auto
    with rstepSub have steps: "\<forall> i. (?lt i \<cdot> ?lsig i, ?ls (Suc i) \<cdot> ?lsig (Suc i)) \<in> (rstep lR)^*" by auto
    
    from chain have mem: "\<forall> i. (s i, t i) \<in> P" by simp
    have "\<forall> i. (?ls i, ?lt i) \<in> ?lP"
    proof
      fix i
      have memI: "(s i, t i) \<in> P" using mem by simp
      have "wf_ass (sub_ass default_ass (\<sigma> i))" 
	by (rule wf_ass_sub_ass, rule wf_default_ass)
      hence "(?ls i, ?lt i) \<in> lab_rule I L C (s i, t i)"
	unfolding lab_rule_def by (auto)
      with memI show "(?ls i, ?lt i) \<in> ?lP" by auto
    qed
    hence memLab: "\<forall> i. (?ls i, ?lt i) \<in> lP" using lP by auto
    
    have "remove_lab_trs lR = remove_lab_trs (unlab_trs lR)" by (simp only: remove_label_trs_2)
    also have "\<dots> \<subseteq> remove_lab_trs R" by (rule map_trs_mono, rule lR2)
    also have "\<dots> = remove_lab_trs ?lR" by (simp only: remove_label_trs)
    finally have lrSub: "rstep (remove_lab_trs lR) \<subseteq> rstep (remove_lab_trs ?lR)" (is "?R1 \<subseteq> ?R2") by (rule rstep_mono)
    from chain have "\<forall> i. SN (rstep (remove_lab_trs R)) (remove_lab_term (t i \<cdot> \<sigma> i))" by (auto simp: lab_minimal_cond_def)
    hence "\<forall> i. SN (rstep (remove_lab_trs (?lR))) (remove_lab_term (?lts i))" by (simp only: remove_label_trs remove_label_term, auto)    
    hence minimal: "\<forall> i. SN (rstep (remove_lab_trs lR)) (remove_lab_term (?lts i))" using SN_imp_SN_subset[where ?R' = ?R1 and ?rR' = ?R2] lrSub by blast 
    
    from steps minimal memLab have "min_lab_ichain (lP,lR) ?ls ?lt ?lsig" by (auto simp: lab_minimal_cond_def)
    with finite show False unfolding finite_lab_dpp_def by blast
  qed
qed

lemma sem_lab_trs_sound : fixes R :: "(('f,'l)lab,'v)trel"
  assumes sem_lab: "sem_lab_trs R lR" and SN: "SN (rstep lR)"
  shows "SN (rstep R)"
unfolding SN_ars_def
proof (intro allI)
  fix t :: "(('f,'l)lab,'v)term"
  show "SN (rstep R) t" 
  proof (rule  SN_lab_imp_SN_term)
    from assms show "model I L C R" unfolding sem_lab_trs_def by auto
  next
    show "SN (rstep (Lab_trs R)) (Lab default_ass t)" 
    proof (rule SN_imp_SN_subset)
      show "rstep (Lab_trs R) \<subseteq> rstep lR" using assms unfolding sem_lab_trs_def using rstep_mono  by blast
    next
      show "SN (rstep lR) (Lab default_ass t)" using SN unfolding SN_ars_def by auto
    qed
  qed
qed
end

datatype arithFun = Arg nat | Const nat | Sum "arithFun list" | Max "arithFun list" | Min "arithFun list" | Prod "arithFun list" 

instantiation arithFun :: sshow
begin
fun sshow_arithFun :: "arithFun \<Rightarrow> sstring"
where "sshow_arithFun (Arg i) = (''x_'' +++ (sshow (Suc i)))"
   |  "sshow_arithFun (Const n) = sshow n"
   |  "sshow_arithFun (Sum (f # fs)) = (sshow f +++ sConcat (map (\<lambda> g. '' + '' +++ sshow g) fs))"
   |  "sshow_arithFun (Prod (f # fs)) = (''('' +++ sshow f +++ sConcat (map (\<lambda> g. '' * '' +++ sshow g) fs) +++ '')'')"
   |  "sshow_arithFun (Min (f # fs)) = (''min('' +++ sshow f +++ sConcat (map (\<lambda> g. '','' +++ sshow g) fs) +++ '')'')"
   |  "sshow_arithFun (Max (f # fs)) = (''max('' +++ sshow f +++ sConcat (map (\<lambda> g. '','' +++ sshow g) fs) +++ '')'')"
instance ..
end

fun take_default where "take_default def [] _ = def" | "take_default _ (x # xs) 0 = x" | "take_default def (_ # xs) (Suc i) = take_default def xs i"

fun eval_arithFun_unbound :: "nat list \<Rightarrow> arithFun \<Rightarrow> nat"
where "eval_arithFun_unbound nats (Arg i) = take_default 0 nats i"
    | "eval_arithFun_unbound nats (Const n) = n"
    | "eval_arithFun_unbound nats (Sum []) = 0"
    | "eval_arithFun_unbound nats (Sum (f # fs)) =  (eval_arithFun_unbound nats f + eval_arithFun_unbound nats (Sum fs))"
    | "eval_arithFun_unbound nats (Prod []) = 1"
    | "eval_arithFun_unbound nats (Prod (f # fs)) =  (eval_arithFun_unbound nats f * eval_arithFun_unbound nats (Prod fs))"
    | "eval_arithFun_unbound nats (Max [f]) = eval_arithFun_unbound nats f"
    | "eval_arithFun_unbound nats (Max (f # fs)) =  (max (eval_arithFun_unbound nats f) (eval_arithFun_unbound nats (Max fs)))"
    | "eval_arithFun_unbound nats (Min [f]) = eval_arithFun_unbound nats f"
    | "eval_arithFun_unbound nats (Min (f # fs)) =  (min (eval_arithFun_unbound nats f) (eval_arithFun_unbound nats (Min fs)))"

declare eval_arithFun_unbound.simps[simp del]

fun eval_arithFun :: "nat \<Rightarrow> arithFun \<Rightarrow> nat list \<Rightarrow> nat"
where "eval_arithFun c f nats = (eval_arithFun_unbound nats f) mod c"


datatype ('f,'l)sl_inter = SL_Inter nat "((('f,'l)lab  \<times> nat) \<times> arithFun)list"

instantiation lab :: (sshow,sshow)sshow
begin
fun sshow_lab :: "('f :: sshow,'l ::sshow)lab \<Rightarrow> string"
where "sshow_lab (Lab f []) = sshow f"
   |  "sshow_lab (Lab f (l # ls)) = (sshow f +++ ''['' +++ sshow l +++ sConcat (map (\<lambda> la. '','' +++ sshow la) ls) +++ '']'')"
instance ..
end


instantiation sl_inter :: (sshow,sshow)sshow
begin
fun sshow_sl_inter :: "('f :: sshow, 'l :: sshow)sl_inter \<Rightarrow> string" 
where "sshow_sl_inter (SL_Inter n fnas) = (
  showList ((''carrier {0,...,'' +++ sshow n +++ ''}'') #
     (map (\<lambda> ((f,n),a). ''['' +++ sshow f +++ ''/''+++ sshow n+++''] = '' +++ sshow a) fnas)))"
instance ..
end

fun get_largest_element where "get_largest_element (SL_Inter n _) = n"

fun sl_inter_to_inter :: "('f,'l)sl_inter \<Rightarrow> ('f,nat,'l)inter"
where "sl_inter_to_inter (SL_Inter c ls) fl cs = (case lookup (fl,length cs) ls of 
       None \<Rightarrow> 0
     | Some f \<Rightarrow> eval_arithFun (Suc c) f cs)"

lemma wf_sl_inter : 
  shows "wf_inter (sl_inter_to_inter sli) {x | x.  x < Suc (get_largest_element sli)}"
proof (cases sli)
  case (SL_Inter c ls)
  thus ?thesis unfolding wf_inter_def
  proof (intro allI, intro impI)
    fix f cs
    show "sl_inter_to_inter sli f cs \<in> {x | x. x < Suc (get_largest_element sli)}" using SL_Inter
      by (cases "lookup (f, length cs) ls", auto)
  qed
qed

interpretation sl_finite: sl_interpr "{x | x . x < Suc (get_largest_element sli)}" "get_largest_element sli" "sl_inter_to_inter sli" 
  by (unfold_locales, simp, rule wf_sl_inter)


(* now let us head for checking the model condition *)


fun enum_vectors :: "'c list \<Rightarrow> 'v list \<Rightarrow> ('v \<times> 'c)list list"
where "enum_vectors C [] = [[]]"
    | "enum_vectors C (x # xs) = (let rec = enum_vectors C xs in concat (map (\<lambda> vec. map (\<lambda> c. (x,c) # vec) C) rec))" 

fun assign_from_vector :: "('v \<times> 'c)list \<Rightarrow> 'v \<Rightarrow> 'c"
where "assign_from_vector vec x = the (lookup x vec)"

lemma enum_vectors_complete: assumes "C \<noteq> []"
  shows "\<exists> vec \<in> set (enum_vectors C xs). \<forall> x \<in> set xs. \<forall> c \<in> set C. \<alpha> x = c \<longrightarrow> assign_from_vector vec x = c"
proof (induct xs, simp)
  case (Cons x xs)
  from this obtain vec where enum: "vec \<in> set (enum_vectors C xs)" and eq: "\<forall> x \<in> set xs. \<forall> c \<in> set C. \<alpha> x = c \<longrightarrow> assign_from_vector vec x = c" by auto
  from enum have res: "set (map (\<lambda> c. (x,c) # vec) C) \<subseteq> set (enum_vectors C (x # xs)) " (is "_ \<subseteq> ?res") by auto
  show ?case 
  proof (cases "\<alpha> x \<in> set C")
    case False
    from assms have "hd C \<in> set C" by auto
    with res have elem: "(x,hd C) # vec \<in> ?res" (is "?vec \<in> _") by auto
    from eq False have "\<forall> y \<in> set (x # xs). \<forall> c \<in> set C. \<alpha> y = c \<longrightarrow> assign_from_vector ?vec y = c" by auto
    with elem show ?thesis by blast
  next
    case True
    with res have elem: "(x, \<alpha> x) # vec \<in> ?res" (is "?vec \<in> _") by auto
    from eq True have "\<forall> y \<in> set (x # xs). \<forall> c \<in> set C. \<alpha> y = c \<longrightarrow> assign_from_vector ?vec y = c" by auto
    with elem show ?thesis by blast
  qed
qed

declare assign_from_vector.simps[simp del]    

lemma eval_lab_independent: assumes "\<forall> x \<in> vars t. \<alpha> x = \<beta> x"
  shows "eval_lab I L \<alpha> t = eval_lab I L \<beta> t"
using assms
proof (induct t, simp)
  case (Fun f ts)
  hence map: "map (eval_lab I L \<alpha>) ts = map (eval_lab I L \<beta>) ts" by auto
  thus ?case by (auto simp: Let_def map)
qed


fun check_sl_rule_ass :: "bool \<Rightarrow> ('f :: sshow,'c :: sshow,'l :: sshow)inter \<Rightarrow> ('f,'c,'l)label \<Rightarrow> (('f,'l)lab,'v)trsL \<Rightarrow> ('v :: sshow,'c)assign \<Rightarrow> (('f,'l)lab,'v)rule \<Rightarrow> string monad"
where "check_sl_rule_ass mc I L lR \<alpha> (l,r) = 
  (let cl_ll = eval_lab I L \<alpha> l in
  (let cr_lr = eval_lab I L \<alpha> r in
    checkB (mc \<longrightarrow> (fst cl_ll = fst cr_lr)) (''rule '' +++ showRule (l,r) +++ '' violates model condition, [lhs] = '' +++ sshow (fst cl_ll) +++ '', [rhs] = '' +++ sshow (fst cr_lr))
  \<guillemotright> checkB ((snd cl_ll, snd cr_lr) mem lR) (''labeled rule '' +++ showRule (snd cl_ll, snd cr_lr) +++ '' missing'')))"

fun check_sl_rule :: "('f :: sshow,'c :: sshow,'l :: sshow)inter \<Rightarrow> ('f,'c,'l)label \<Rightarrow> 'c list \<Rightarrow> bool \<Rightarrow> (('f,'l)lab,'v)trsL \<Rightarrow> (('f,'l)lab,'v :: sshow)rule \<Rightarrow> string monad"
where "check_sl_rule I L C mc lR (l,r) = checkAllM (\<lambda> \<alpha> . check_sl_rule_ass mc I L lR \<alpha> (l,r)) 
             (map assign_from_vector (enum_vectors C (unionI (varsLI l) (varsLI r))))"

lemma check_sl_rule_sound[simp]: fixes lR :: "(('f :: sshow,'l :: sshow)lab,'v :: sshow)trsL" and I :: "('f :: sshow,'c :: sshow,'l :: sshow)inter"
  assumes "isOK (check_sl_rule I L C mc lR (l,r))"
  shows "(mc \<longrightarrow> model_rule I L (set C) l r) \<and> lab_rule I L (set C) (l,r) \<subseteq> set lR"
proof (cases C)
  case Nil
  thus ?thesis unfolding lab_rule_def wf_assign_def by auto
next
  case (Cons c D)
  let ?C = "Cons c D"
  let ?V = "unionI (varsLI l) (varsLI r) :: 'v list"
  from assms Cons have check: "\<forall> \<alpha> \<in> set (map assign_from_vector (enum_vectors ?C ?V)). isOK(check_sl_rule_ass mc I L lR \<alpha> (l,r))"
    by auto
  have "\<forall> \<alpha>. wf_assign (set ?C) \<alpha> \<longrightarrow> isOK(check_sl_rule_ass mc I L lR \<alpha> (l,r))"
  proof (intro allI, intro impI)
    fix \<alpha> :: "'v \<Rightarrow> 'c"
    assume wf_ass: "wf_assign (set ?C) \<alpha>"
    from enum_vectors_complete obtain vec where vec: "vec \<in> set (enum_vectors ?C ?V) \<and> (\<forall> x \<in> set ?V. \<forall> c \<in> set ?C. \<alpha> x = c \<longrightarrow> assign_from_vector vec x = c)"
      by best
    let ?\<beta> = "assign_from_vector vec"
    from wf_ass vec have "\<forall> x \<in> set ?V. \<alpha> x = ?\<beta> x" unfolding wf_assign_def by auto
    hence "(\<forall> x \<in> vars l. ?\<beta> x = \<alpha> x) \<and> (\<forall> x \<in> vars r. ?\<beta> x = \<alpha> x) " by auto
    hence equal: "eval_lab I L \<alpha> l = eval_lab I L ?\<beta> l \<and> eval_lab I L \<alpha> r = eval_lab I L ?\<beta> r" 
      using eval_lab_independent by force
    from vec check have "isOK(check_sl_rule_ass mc I L lR ?\<beta> (l,r))" by auto
    with equal show "isOK(check_sl_rule_ass mc I L lR \<alpha> (l,r))" by auto
  qed
  thus ?thesis using Cons unfolding lab_rule_def by (auto simp: Let_def mem_iff)
qed

declare check_sl_rule.simps[simp del]

types ('f,'c,'l,'v)sl_check1 = "('f,'c,'l)inter \<Rightarrow> ('f,'c,'l)label \<Rightarrow> 'c list \<Rightarrow> (('f,'l)lab,'v)trsL \<Rightarrow> (('f,'l)lab,'v)trsL \<Rightarrow> string monad"
types ('f,'l,'v)sl_check2 = "(('f,'l)lab,'v)trsL \<Rightarrow> (('f,'l)lab,'v)trsL \<Rightarrow> string monad"

definition check_sl_model_lab_trs :: "('f :: sshow, 'c :: sshow, 'l :: sshow, 'v :: sshow)sl_check1" 
  where "check_sl_model_lab_trs I L C lR R \<equiv> checkAllM (check_sl_rule I L C True lR) R"
definition check_sl_lab_trs :: "('f :: sshow, 'c :: sshow, 'l :: sshow, 'v :: sshow)sl_check1" 
  where "check_sl_lab_trs I L C lP P \<equiv> checkAllM (check_sl_rule I L C False lP) P"
definition check_sl_unlab_trs where "check_sl_unlab_trs lR R = checkSubset (map_trsI unlab lR) R"

lemma check_sl_lab_trs_sound[simp]:
  assumes "isOK(check_sl_lab_trs I L C lP P)"
  shows "lab_trs I L (set C) (set P) \<subseteq> set lP"
proof -
  from assms have check: "\<forall> (l,r) \<in> set P. isOK(check_sl_rule I L C False lP (l,r))"
    unfolding check_sl_lab_trs_def by auto
  have "\<forall> (l,r) \<in> set P. lab_rule I L (set C) (l,r) \<subseteq> set lP" 
  proof (clarify)
    fix l r ll lr
    assume lr: "(l,r) \<in> set P" and llr: "(ll,lr) \<in> lab_rule I L (set C) (l,r)"
    from lr check have "isOK(check_sl_rule I L C False lP (l,r))" by auto 
    hence "lab_rule I L (set C) (l,r) \<subseteq> set lP" by auto
    with lr llr show "(ll,lr) \<in> set lP" by auto
  qed
  thus ?thesis by auto
qed


lemma check_sl_model_lab_trs_sound[simp]:
  assumes "isOK(check_sl_model_lab_trs I L C lR R)"
  shows "lab_trs I L (set C) (set R) \<subseteq> set lR \<and> model I L (set C) (set R)"
proof -
  from assms have check: "\<forall> (l,r) \<in> set R. isOK(check_sl_rule I L C True lR (l,r))"
    unfolding check_sl_model_lab_trs_def by auto
  have "\<forall> (l,r) \<in> set R. model_rule I L (set C) l r \<and> lab_rule I L (set C) (l,r) \<subseteq> set lR" 
  proof (clarify)
    fix l r
    assume lr: "(l,r) \<in> set R" 
    from lr check have "isOK(check_sl_rule I L C True lR (l,r))" by auto 
    thus "model_rule I L (set C) l r \<and> lab_rule I L (set C) (l,r) \<subseteq> set lR" by auto
  qed
  thus ?thesis unfolding model_def by auto
qed

definition check_sem_lab_proc :: "('f :: sshow, 'l :: sshow, 'v :: sshow)sl_check2
  \<Rightarrow> ('f,'l,'v)sl_check2 \<Rightarrow> ('f,'c,'l)inter \<Rightarrow> ('f,'c,'l)label \<Rightarrow> 
  (('f,'l)lab,'v)trsL \<Rightarrow> 
  (('f,'l)lab,'v)trsL \<Rightarrow> 
  (('f,'l)lab,'v)trsL \<Rightarrow> 
  (('f,'l)lab,'v)trsL \<Rightarrow> 
  string list monad"
where "check_sem_lab_proc check_lab check_model_lab I L P R lP lR \<equiv> 
  check_lab lP P
    \<leftarrow> (\<lambda> s. ''not all labeled DPs present:'' # [s])
\<guillemotright> check_model_lab lR R
    \<leftarrow> (\<lambda> s. ''problem with labeled TRS:'' # [s])
\<guillemotright> check_sl_unlab_trs lR R
    \<leftarrow> (\<lambda> lr. ''problem with labeled TRS:'' # [''for the labeled rule '' +++ showRule lr +++ '' there is no unlabeled counterpart''])"

definition check_sem_lab_trs :: "('f :: sshow, 'l :: sshow, 'v :: sshow)sl_check2
  \<Rightarrow> ('f,'l,'v)sl_check2 \<Rightarrow> ('f,'c,'l)inter \<Rightarrow> ('f,'c,'l)label \<Rightarrow> 
  (('f,'l)lab,'v)trsL \<Rightarrow> 
  (('f,'l)lab,'v)trsL \<Rightarrow> 
  string list monad"
where "check_sem_lab_trs check_lab check_model_lab I L R lR \<equiv> 
  check_model_lab lR R
    \<leftarrow> (\<lambda> s. ''problem with labeled TRS:'' # [s])"

locale sl_interpr_impl = sl_interpr +
  fixes check_lab :: "('b,'c,'d :: sshow)sl_check2"
   and  check_model_lab :: "('b,'c,'d)sl_check2"
 assumes check_lab: "isOK(check_lab lR R) \<Longrightarrow> lab_trs I L C (set R) \<subseteq> set lR"
   and   check_model_lab: "isOK(check_model_lab lR R) \<Longrightarrow> lab_trs I L C (set R) \<subseteq> set lR \<and> model I L C (set R)"
begin 

lemma check_sem_lab_proc_sound : 
  "sound_lab_proc_impl (\<lambda> P R rm lP lR. isOK(check_sem_lab_proc check_lab check_model_lab I L P R lP lR))"
unfolding sound_lab_proc_impl_def
proof (intro allI, intro impI)
  fix P R lP lR 

  assume check: "isOK(check_sem_lab_proc check_lab check_model_lab I L P R lP lR)"
    and  finite: "finite_lab_dpp (set lP, set lR)"
  hence "Lab_trs (set P) \<subseteq> set lP \<and> Lab_trs (set R) \<subseteq> set lR \<and> model I L C (set R) \<and> unlab_trs (set lR) \<subseteq> set R" using check_lab check_model_lab 
    unfolding check_sem_lab_proc_def by (auto simp: check_sl_unlab_trs_def)
  hence "sem_lab_proc (set P) (set R) (set lP) (set lR)" unfolding sem_lab_proc_def by auto
  with finite sem_lab_proc_sound show "finite_lab_dpp (set P, set R)" unfolding sound_lab_proc_def by blast
qed

lemma check_sem_lab_trs_sound : 
  assumes check: "isOK(check_sem_lab_trs check_lab check_model_lab I L R lR)"
  and SN: "SN (rstep (set lR))"
  shows "SN (rstep (set R))"
proof (rule sem_lab_trs_sound)
  from check show "sem_lab_trs (set R) (set lR)" unfolding check_sem_lab_trs_def sem_lab_trs_def using check_model_lab by (simp add: check_model_lab)
next
  from SN show "SN (rstep (set lR))" .
qed
end

lemma sl_helper:  "{x | x. x < Suc (get_largest_element sli)} = set [0..<Suc (get_largest_element sli)]" by auto

interpretation sl_finite_impl : sl_interpr_impl "{x | x . x < Suc (get_largest_element sli)}" "get_largest_element sli" "sl_inter_to_inter sli" L  
  "check_sl_lab_trs (sl_inter_to_inter sli) L [0 ..< Suc (get_largest_element sli)]"
  "check_sl_model_lab_trs (sl_inter_to_inter sli) L [0 ..< Suc (get_largest_element sli)]" 
  by (unfold_locales, simp only: sl_helper, simp only: check_sl_lab_trs_sound,
     simp only: sl_helper, simp only: check_sl_model_lab_trs_sound, auto)

definition create_sem_lab_trs  :: "
  ('f :: sshow,'l :: sshow)sl_inter \<Rightarrow> ('f,nat,'l)label \<Rightarrow> 
  (('f,'l)lab,'v :: sshow)trsL \<Rightarrow> 
  (('f,'l)lab,'v)trsL \<Rightarrow> 
  string list monad" 
where "create_sem_lab_trs sli L r lr \<equiv> 
  let c = get_largest_element sli in (
  let C = [0 ..< Suc c] in (
  let I = sl_inter_to_inter sli in (
  (check_sem_lab_trs
    (check_sl_lab_trs I L C) 
    (check_sl_model_lab_trs I L C) I L r lr)
   \<leftarrow> (\<lambda> s. sshow sli # s)
)))"


definition create_sem_lab_proc  :: "
  ('f :: sshow,'l :: sshow)sl_inter \<Rightarrow> ('f,nat,'l)label \<Rightarrow> 
  (('f,'l)lab,'v :: sshow)trsL \<Rightarrow> 
  (('f,'l)lab,'v)trsL \<Rightarrow> 
  (('f,'l)lab,'v)trsL \<Rightarrow> 
  (('f,'l)lab,'v)trsL \<Rightarrow> 
  string list monad" 
where "create_sem_lab_proc sli L r1 r2 r3 r4 \<equiv> 
  let c = get_largest_element sli in (
  let C = [0 ..< Suc c] in (
  let I = sl_inter_to_inter sli in (
  (check_sem_lab_proc 
    (check_sl_lab_trs I L C) 
    (check_sl_model_lab_trs I L C) I L r1 r2 r3 r4)
   \<leftarrow> (\<lambda> s. sshow sli # s))))"

lemma create_sem_lab_proc_sound : "sound_lab_proc_impl (\<lambda> P R rm lP lR. isOK(create_sem_lab_proc sli L P R lP lR))"
using sl_finite_impl.check_sem_lab_proc_sound 
  by (simp add: Let_def create_sem_lab_proc_def, blast)

lemma create_sem_lab_trs_sound : assumes "isOK(create_sem_lab_trs sli L R lR)"
  and "SN (rstep (set lR))"
  shows "SN (rstep (set R))"
using sl_finite_impl.check_sem_lab_trs_sound assms
  by (simp add: Let_def create_sem_lab_trs_def, blast)

definition check_unlab_trs :: 
  "(('f :: sshow,'l :: sshow)lab,'v :: sshow)trsL \<Rightarrow> 
   (('f,'l)lab,'v)trsL \<Rightarrow> 
   string list monad" 
where "check_unlab_trs R uR \<equiv> checkSubset (map_trsI unlab R) uR
   \<leftarrow> (\<lambda> rule. [''rule '' +++ showRule rule +++ '' is missing in unlabeled TRS''])"

lemma check_unlab_trs_sound : fixes R :: "(('f :: sshow,'l :: sshow)lab,'v :: sshow)trsL"
  assumes check: "isOK(check_unlab_trs R uR)"
  and SN: "SN (rstep (set uR))"
  shows "SN (rstep (set R))"
unfolding SN_ars_def
proof (rule allI)
  fix t :: "(('f,'l)lab,'v)term"
  from check have subset: "unlab_trs (set R) \<subseteq> set uR" unfolding check_unlab_trs_def by auto
  have SN2: "SN (rstep (unlab_trs (set R))) (unlab_term t)"
  proof (rule SN_imp_SN_subset)
    show "SN (rstep (set uR)) (unlab_term t)" using SN unfolding SN_ars_def by auto
  next
    show "rstep (unlab_trs (set R)) \<subseteq> rstep (set uR)" 
      by (rule rstep_mono, rule subset)
  qed
  show "SN (rstep (set R)) t"
  proof (rule SN_map_imp_SN)
    show "SN (rstep (unlab_trs (set R))) (unlab_term t)" using SN2  .
  qed
qed

end